(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.logger = factory());
}(this, (function () { 'use strict';

  function _slicedToArray(arr, i) {
    return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
  }

  function _arrayWithHoles(arr) {
    if (Array.isArray(arr)) return arr;
  }

  function _iterableToArrayLimit(arr, i) {
    var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"];

    if (_i == null) return;
    var _arr = [];
    var _n = true;
    var _d = false;

    var _s, _e;

    try {
      for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"] != null) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  function _unsupportedIterableToArray(o, minLen) {
    if (!o) return;
    if (typeof o === "string") return _arrayLikeToArray(o, minLen);
    var n = Object.prototype.toString.call(o).slice(8, -1);
    if (n === "Object" && o.constructor) n = o.constructor.name;
    if (n === "Map" || n === "Set") return Array.from(o);
    if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
  }

  function _arrayLikeToArray(arr, len) {
    if (len == null || len > arr.length) len = arr.length;

    for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

    return arr2;
  }

  function _nonIterableRest() {
    throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var assign = Object.assign;

  var humanize = function (ms) {
    if (ms >= 1000) {
      return "".concat(ms / 1000, "s");
    }

    return "".concat(ms, "ms");
  };

  var stringToRGB = function (string) {
    var code = string.split('').reduce(function (hash, _, i) {
      return "".concat(string.charCodeAt(i) + (hash << 5) - hash);
    }, 0);
    var hex = (code & 0x00FFFFFF).toString(16);
    return "#".concat('00000'.slice(0, 6 - hex.length) + hex);
  };

  var cloneTree = function (tree) {
    return assign({}, tree, {
      attributes: assign({}, tree.attributes),
      childNodes: tree.childNodes.map(function (vTree) {
        return cloneTree(vTree);
      })
    });
  };

  var Internals = null;
  /**
   * @returns {any}
   */

  var format = function (patches) {
    if (!Internals) {
      return {};
    }

    var _Internals2 = Internals,
        decodeEntities = _Internals2.decodeEntities,
        PATCH_TYPE = _Internals2.PATCH_TYPE;
    var newPatches = {
      'Node (Insert)': [],
      'Node (Remove)': [],
      'Node (Replace)': [],
      'Node value (Set)': [],
      'Attribute (Set)': [],
      'Attribute (Remove)': []
    };
    var length = patches.length;
    var i = 0;
    var uniquePatches = 0;

    while (true) {
      var patchType = patches[i];

      if (i === length) {
        break;
      }

      uniquePatches += 1;

      switch (patchType) {
        case PATCH_TYPE.SET_ATTRIBUTE:
          {
            var vTree = patches[i + 1];
            var name = patches[i + 2];
            var value = decodeEntities(patches[i + 3]);
            newPatches['Attribute (Set)'].push({
              'Node': Internals.NodeCache.get(vTree),
              'Name': name,
              'Value': value
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.REMOVE_ATTRIBUTE:
          {
            var _vTree = patches[i + 1];
            var _name = patches[i + 2];
            newPatches['Attribute (Remove)'].push({
              'Node': Internals.NodeCache.get(_vTree),
              'Name': _name
            });
            i += 3;
            break;
          }

        case PATCH_TYPE.NODE_VALUE:
          {
            var _vTree2 = patches[i + 1];
            var nodeValue = patches[i + 2];
            var oldValue = patches[i + 3];
            newPatches['Node value (Set)'].push({
              'Node': Internals.NodeCache.get(_vTree2),
              'Value': nodeValue,
              'Old value': oldValue
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.INSERT_BEFORE:
          {
            var _vTree3 = patches[i + 1];
            var newTree = patches[i + 2];
            var refTree = patches[i + 3];
            newPatches['Node (Insert)'].push({
              'Container': Internals.NodeCache.get(_vTree3),
              'Node': Internals.NodeCache.get(newTree),
              'Insert Before': Internals.NodeCache.get(refTree)
            });
            i += 4;
            break;
          }

        case PATCH_TYPE.REPLACE_CHILD:
          {
            var _newTree = patches[i + 1];
            var oldTree = patches[i + 2];
            newPatches['Node (Replace)'].push({
              'New node': Internals.NodeCache.get(_newTree),
              'New tree': _newTree,
              'Old tree': oldTree
            });
            i += 3;
            break;
          }

        case PATCH_TYPE.REMOVE_CHILD:
          {
            var _vTree4 = patches[i + 1];
            newPatches['Node (Remove)'].push({
              'Tree': _vTree4
            });
            i += 2;
            break;
          }
      }
    }

    return [newPatches, uniquePatches];
  };
  /**
   * Re-usable log function. Used for during render and after render.
   *
   * @param {string} message - Prefix for the console output.
   * @param {string} method - Which console method to call
   * @param {string} color - Which color styles to use
   * @param {Date} _date - A date object to render
   * @param {import('diffhtml/lib/transaction').default} transaction - Contains: domNode, oldTree, newTree, patches, promises
   * @param {any=} completed - Middleware options
   */


  var log = function (message, method, color, _date, transaction, completed) {
    var mount = transaction.mount,
        newTree = transaction.newTree,
        patches = transaction.patches,
        promises = transaction.promises,
        config = transaction.config,
        input = transaction.input,
        state = transaction.state; // Shadow DOM rendering...

    if (mount.host) {
      var ctorName = mount.host.constructor.name;
      console[method]("%c".concat(ctorName, " render ").concat(completed ? 'ended' : 'started'), "color: ".concat(stringToRGB(ctorName)).concat(completed ? '; opacity: 0.5' : ''), completed ? completed : '');
    } else {
      console[method](message, color, completed ? completed : '');
    }

    if (!completed && mount) {
      console.log('%cmount %O', 'font-weight: bold; color: #868686', mount);
    }

    if (!completed && input) {
      console.log('%cmarkup %O', 'font-weight: bold; color: #868686', input);
    }

    if (!completed && config) {
      console.log('%coptions', 'font-weight: bold; color: #868686', config);
    }

    if (state) {
      console.log('%cstate %O', 'font-weight: bold; color: #868686', state);
    }

    if (transaction._cloneOldTree || newTree) {
      console.log('%coldTree %O newTree %O', 'font-weight: bold; color: #868686', transaction._cloneOldTree, cloneTree(newTree));
    }

    if (patches && patches.length) {
      var _format = format(patches),
          _format2 = _slicedToArray(_format, 2),
          formattedPatches = _format2[0],
          uniqueCount = _format2[1];

      console.log('%cpatches (%d) %O', 'font-weight: bold; color: #868686', uniqueCount, formattedPatches);
    } // Don't clutter the output if there aren't any promises.


    if (promises && promises.length) {
      console.log('%ctransition promises %O', 'font-weight: bold; color: #868686', promises);
    }
  };

  var logger = function (_ref) {
    var _ref$minimize = _ref.minimize,
        minimize = _ref$minimize === void 0 ? false : _ref$minimize;
    return assign(function loggerTask(transaction) {
      var start = new Date();
      var mount = transaction.mount;
      var name = mount.name || mount.displayName || mount.nodeName;
      log("%c<".concat(transaction.mount.nodeName, "> render started  "), 'groupCollapsed', 'color: #FF0066', start, transaction);
      var oldTree = transaction.oldTree;

      if (transaction.state.isRendering) {
        console.groupEnd();
        log("%c<".concat(name, "> render aborted  "), minimize ? 'groupCollapsed' : 'group', 'color: #FF78B2', new Date(), transaction);
        console.groupEnd();
        return;
      }

      transaction._cloneOldTree = oldTree && cloneTree(oldTree);
      /**
       * Transaction is effectively done, but we need to listen for it to actually
       * be finished.
       */

      return function () {
        // Transaction has fully completed.
        transaction.onceEnded(function () {
          console.groupEnd();
          log("%c<".concat(name, "> render ended  "), minimize ? 'groupCollapsed' : 'group', 'color: #FF78B2', new Date(), transaction, ' >> Completed in: ' + humanize(Date.now() - Number(start)));
          console.groupEnd();
        });
      };
    }, {
      displayName: 'loggerTask',
      subscribe: function (_Internals) {
        Internals = _Internals;
      }
    });
  };

  var index = (function () {
    var opts = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    return logger(opts);
  });

  return index;

})));
